"""All concepts related to threats."""

import chess

from lczerolens.board import LczeroBoard
from lczerolens.concept import BinaryConcept


class HasThreat(BinaryConcept):
    """
    Class for material concept-based XAI methods.
    """

    def __init__(
        self,
        piece: str,
        relative: bool = True,
    ):
        """
        Initialize the class.
        """
        self.piece = chess.Piece.from_symbol(piece)
        self.relative = relative

    def compute_label(
        self,
        board: LczeroBoard,
    ) -> int:
        """
        Compute the label for a given model and input.
        """
        if self.relative:
            color = self.piece.color if board.turn else not self.piece.color
        else:
            color = self.piece.color
        squares = board.pieces(self.piece.piece_type, color)
        for square in squares:
            if board.is_attacked_by(not color, square):
                return 1
        return 0


class HasMateThreat(BinaryConcept):
    """
    Class for material concept-based XAI methods.
    """

    def compute_label(
        self,
        board: LczeroBoard,
    ) -> int:
        """
        Compute the label for a given model and input.
        """
        for move in board.legal_moves:
            board.push(move)
            if board.is_checkmate():
                board.pop()
                return 1
            board.pop()
        return 0



"""All concepts related to material."""

from typing import Dict, Optional

import chess

from lczerolens.board import LczeroBoard
from lczerolens.concept import BinaryConcept


class HasPiece(BinaryConcept):
    """Class for material concept-based XAI methods."""

    def __init__(
        self,
        piece: str,
        relative: bool = True,
    ):
        """Initialize the class."""
        self.piece = chess.Piece.from_symbol(piece)
        self.relative = relative

    def compute_label(
        self,
        board: LczeroBoard,
    ) -> int:
        """Compute the label for a given model and input."""
        if self.relative:
            color = self.piece.color if board.turn else not self.piece.color
        else:
            color = self.piece.color
        squares = board.pieces(self.piece.piece_type, color)
        return 1 if len(squares) > 0 else 0


class HasMaterialAdvantage(BinaryConcept):
    """Class for material concept-based XAI methods.

    Attributes
    ----------
    piece_values : Dict[int, int]
        The piece values.
    """

    piece_values = {
        chess.PAWN: 1,
        chess.KNIGHT: 3,
        chess.BISHOP: 3,
        chess.ROOK: 5,
        chess.QUEEN: 9,
        chess.KING: 0,
    }

    def __init__(
        self,
        relative: bool = True,
    ):
        """
        Initialize the class.
        """
        self.relative = relative

    def compute_label(
        self,
        board: LczeroBoard,
        piece_values: Optional[Dict[int, int]] = None,
    ) -> int:
        """
        Compute the label for a given model and input.
        """
        if piece_values is None:
            piece_values = self.piece_values
        if self.relative:
            us, them = board.turn, not board.turn
        else:
            us, them = chess.WHITE, chess.BLACK
        our_value = 0
        their_value = 0
        for piece in range(1, 7):
            our_value += len(board.pieces(piece, us)) * piece_values[piece]
            their_value += len(board.pieces(piece, them)) * piece_values[piece]
        return 1 if our_value > their_value else 0



"""Utils from the lczero executable and bindings.

Notes
----
The lczero bindings are not installed by default. You can install them by
running `pip install lczero`.
"""

import subprocess

import chess
import torch

from lczerolens.board import LczeroBoard

try:
    from lczero.backends import Backend, GameState
except ImportError as e:
    raise ImportError("LCZero bindings are not installed.See https://github.com/LeelaChessZero/lc0.git.") from e


def generic_command(args, verbose=False):
    """
    Run a generic command.
    """
    popen = subprocess.Popen(
        ["lc0", *args],
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
    )
    popen.wait()
    if popen.returncode != 0:
        if verbose:
            stderr = f"\n[DEBUG] stderr:\n{popen.stderr.read().decode('utf-8')}"
        else:
            stderr = ""
        raise RuntimeError(f"Could not run `lc0 {' '.join(args)}`." + stderr)
    return popen.stdout.read().decode("utf-8")


def describenet(path, verbose=False):
    """
    Describe the net at the given path.
    """
    return generic_command(["describenet", "-w", path], verbose=verbose)


def convert_to_onnx(in_path, out_path, verbose=False):
    """
    Convert the net at the given path.
    """
    return generic_command(
        ["leela2onnx", f"--input={in_path}", f"--output={out_path}"],
        verbose=verbose,
    )


def convert_to_leela(in_path, out_path, verbose=False):
    """
    Convert the net at the given path.
    """
    return generic_command(
        ["onnx2leela", f"--input={in_path}", f"--output={out_path}"],
        verbose=verbose,
    )


def board_from_backend(lczero_backend: Backend, lczero_game: GameState, planes: int = 112):
    """
    Create a board from the lczero backend.
    """
    lczero_input = lczero_game.as_input(lczero_backend)
    lczero_input_tensor = torch.zeros((112, 64), dtype=torch.float)
    for plane in range(planes):
        mask_str = f"{lczero_input.mask(plane):b}".zfill(64)
        lczero_input_tensor[plane] = torch.tensor(
            tuple(map(int, reversed(mask_str))), dtype=torch.float
        ) * lczero_input.val(plane)
    return lczero_input_tensor.view((112, 8, 8))


def prediction_from_backend(
    lczero_backend: Backend,
    lczero_game: GameState,
    softmax: bool = False,
    only_legal: bool = False,
    illegal_value: float = 0,
):
    """
    Predicts the move.
    """
    filtered_policy = torch.full((1858,), illegal_value, dtype=torch.float)
    lczero_input = lczero_game.as_input(lczero_backend)
    (lczero_output,) = lczero_backend.evaluate(lczero_input)
    if only_legal:
        indices = torch.tensor(lczero_game.policy_indices())
    else:
        indices = torch.tensor(range(1858))
    if softmax:
        policy = torch.tensor(lczero_output.p_softmax(*range(1858)), dtype=torch.float)
    else:
        policy = torch.tensor(lczero_output.p_raw(*range(1858)), dtype=torch.float)
    value = torch.tensor(lczero_output.q())
    filtered_policy[indices] = policy[indices]
    return filtered_policy, value


def moves_with_castling_swap(lczero_game: GameState, board: LczeroBoard):
    """
    Get the moves with castling swap.
    """
    lczero_legal_moves = lczero_game.moves()
    lczero_policy_indices = list(lczero_game.policy_indices())
    for move in board.legal_moves:
        uci_move = move.uci()
        if uci_move in lczero_legal_moves:
            continue
        if board.is_castling(move):
            leela_uci_move = uci_move.replace("g", "h").replace("c", "a")
            if leela_uci_move in lczero_legal_moves:
                lczero_legal_moves.remove(leela_uci_move)
                lczero_legal_moves.append(uci_move)
                lczero_policy_indices.remove(
                    LczeroBoard.encode_move(
                        chess.Move.from_uci(leela_uci_move),
                        board.turn,
                    )
                )
                lczero_policy_indices.append(LczeroBoard.encode_move(move, board.turn))
    return lczero_legal_moves, lczero_policy_indices



"""Activation lens."""

from lczerolens.model import LczeroModel
from lczerolens.lens import Lens


@Lens.register("activation")
class ActivationLens(Lens):
    """
    Class for activation-based XAI methods.

    Examples
    --------

        .. code-block:: python

            model = LczeroModel.from_path(model_path)
            lens = ActivationLens()
            board = LczeroBoard()
            results = lens.analyse(board, model=model)
    """

    def _intervene(
        self,
        model: LczeroModel,
        **kwargs,
    ) -> dict:
        save_inputs = kwargs.get("save_inputs", False)
        results = {}
        for name, module in self._get_modules(model):
            if save_inputs:
                results[f"{name}_input"] = module.input.save()
            results[f"{name}_output"] = module.output.save()
        return results



"""Generic lens class."""

from abc import ABC, abstractmethod
from typing import Dict, Iterable, Generator, Callable, Type, Union, Optional, Any

import torch
import re

from lczerolens.model import LczeroModel
from lczerolens.board import LczeroBoard


class Lens(ABC):
    """Generic lens class for analysing model activations."""

    _lens_type: str
    _registry: Dict[str, Type["Lens"]] = {}

    @classmethod
    def register(cls, name: str) -> Callable:
        """Registers the lens.

        Parameters
        ----------
        name : str
            The name of the lens.

        Returns
        -------
        Callable
            The decorator to register the lens.

        Raises
        ------
        ValueError
            If the lens name is already registered.
        """

        if name in cls._registry:
            raise ValueError(f"Lens {name} already registered.")

        def decorator(subclass: Type["Lens"]):
            subclass._lens_type = name
            cls._registry[name] = subclass
            return subclass

        return decorator

    @classmethod
    def from_name(cls, name: str, *args, **kwargs) -> "Lens":
        """Returns the lens from its name.

        Parameters
        ----------
        name : str
            The name of the lens.

        Returns
        -------
        Lens
            The lens instance.

        Raises
        ------
        KeyError
            If the lens name is not found.
        """
        if name not in cls._registry:
            raise KeyError(f"Lens {name} not found.")
        return cls._registry[name](*args, **kwargs)

    def __init__(self, pattern: Optional[str] = None):
        """Initialise the lens.

        Parameters
        ----------
        pattern : Optional[str], default=None
            The pattern to match the modules.
        """
        if pattern is None:
            pattern = r"a^"  # match nothing by default
        self._pattern = pattern
        self._reg_exp = re.compile(pattern)

    @property
    def pattern(self) -> str:
        """The pattern to match the modules."""
        return self._pattern

    @pattern.setter
    def pattern(self, pattern: str):
        self._pattern = pattern
        self._reg_exp = re.compile(pattern)

    def _get_modules(self, model: LczeroModel) -> Generator[tuple[str, Any], None, None]:
        """Get the modules to intervene on."""
        for name, module in model.named_modules():
            fixed_name = name.lstrip(". ")  # nnsight outputs names with a dot
            if self._reg_exp.match(fixed_name):
                yield fixed_name, module

    def is_compatible(self, model: LczeroModel) -> bool:
        """Returns whether the lens is compatible with the model.

        Parameters
        ----------
        model : LczeroModel
            The NNsight model.

        Returns
        -------
        bool
            Whether the lens is compatible with the model.
        """
        return isinstance(model, LczeroModel)

    def prepare(self, model: LczeroModel, **kwargs) -> LczeroModel:
        """Prepare the model for the lens.

        Parameters
        ----------
        model : LczeroModel
            The NNsight model.

        Returns
        -------
        LczeroModel
            The prepared model.
        """
        return model

    @abstractmethod
    def _intervene(self, model: LczeroModel, **kwargs) -> dict:
        """Intervene on the model.

        Parameters
        ----------
        model : LczeroModel
            The NNsight model.

        Returns
        -------
        dict
            The intervention results.
        """
        pass

    def _trace(
        self,
        model: LczeroModel,
        *inputs: Union[LczeroBoard, torch.Tensor],
        model_kwargs: dict,
        intervention_kwargs: dict,
    ):
        """Trace the model and intervene on it.

        Parameters
        ----------
        model : LczeroModel
            The NNsight model.
        inputs : Union[LczeroBoard, torch.Tensor]
            The inputs.
        model_kwargs : dict
            The model kwargs.
        intervention_kwargs : dict
            The intervention kwargs.

        Returns
        -------
        dict
            The intervention results.
        """
        with model.trace(*inputs, **model_kwargs):
            return self._intervene(model, **intervention_kwargs)

    def analyse(
        self,
        model: LczeroModel,
        *inputs: Union[LczeroBoard, torch.Tensor],
        **kwargs,
    ) -> dict:
        """Analyse the input.

        Parameters
        ----------
        model : LczeroModel
            The NNsight model.
        inputs : Union[LczeroBoard, torch.Tensor]
            The inputs.

        Returns
        -------
        dict
            The analysis results.

        Raises
        ------
        ValueError
            If the lens is not compatible with the model.
        """
        if not self.is_compatible(model):
            raise ValueError(f"Lens {self._lens_type} is not compatible with the model.")
        model_kwargs = kwargs.get("model_kwargs", {})
        prepared_model = self.prepare(model, **kwargs)
        return self._trace(prepared_model, *inputs, model_kwargs=model_kwargs, intervention_kwargs=kwargs)

    def analyse_batched(
        self,
        model: LczeroModel,
        iter_inputs: Iterable[Union[LczeroBoard, torch.Tensor]],
        **kwargs,
    ) -> Generator[dict, None, None]:
        """Analyse a batches of inputs.

        Parameters
        ----------
        model : LczeroModel
            The NNsight model.
        iter_inputs : Iterable[Tuple[Union[LczeroBoard, torch.Tensor], dict]]
            The iterator over the inputs.

        Returns
        -------
        Generator[dict, None, None]
            The iterator over the statistics.

        Raises
        ------
        ValueError
            If the lens is not compatible with the model.
        """
        if not self.is_compatible(model):
            raise ValueError(f"Lens {self._lens_type} is not compatible with the model.")
        model_kwargs = kwargs.get("model_kwargs", {})
        prepared_model = self.prepare(model, **kwargs)
        for inputs, dynamic_intervention_kwargs in iter_inputs:
            kwargs.update(dynamic_intervention_kwargs)
            yield self._trace(prepared_model, *inputs, model_kwargs=model_kwargs, intervention_kwargs=kwargs)



"""Class for wrapping the LCZero models."""

import os
from typing import Dict, Type, Any, Tuple, Union

import torch
from onnx2torch import convert
from onnx2torch.utils.safe_shape_inference import safe_shape_inference
from tensordict import TensorDict
from torch import nn
from nnsight import NNsight
from contextlib import contextmanager

from lczerolens.board import InputEncoding, LczeroBoard


class LczeroModel(NNsight):
    """Class for wrapping the LCZero models."""

    def trace(
        self,
        *inputs: Any,
        **kwargs: Dict[str, Any],
    ):
        kwargs["scan"] = False
        kwargs["validate"] = False
        return super().trace(*inputs, **kwargs)

    def _execute(self, *prepared_inputs: torch.Tensor, **kwargs) -> Any:
        kwargs.pop("input_encoding", None)
        kwargs.pop("input_requires_grad", None)
        with self._ensure_proper_forward():
            return super()._execute(*prepared_inputs, **kwargs)

    def _prepare_inputs(self, *inputs: Union[LczeroBoard, torch.Tensor], **kwargs) -> Tuple[Tuple[Any], int]:
        input_encoding = kwargs.pop("input_encoding", InputEncoding.INPUT_CLASSICAL_112_PLANE)
        input_requires_grad = kwargs.pop("input_requires_grad", False)

        if len(inputs) == 1 and isinstance(inputs[0], torch.Tensor):
            return inputs, len(inputs[0])
        for board in inputs:
            if not isinstance(board, LczeroBoard):
                raise ValueError(f"Got invalid input type {type(board)}.")

        tensor_list = [board.to_input_tensor(input_encoding=input_encoding).unsqueeze(0) for board in inputs]
        batched_tensor = torch.cat(tensor_list, dim=0)
        if input_requires_grad:
            batched_tensor.requires_grad = True
        batched_tensor = batched_tensor.to(self.device)

        return (batched_tensor,), len(inputs)

    def __call__(self, *inputs, **kwargs):
        prepared_inputs, _ = self._prepare_inputs(*inputs, **kwargs)
        return self._execute(*prepared_inputs, **kwargs)

    def __getattr__(self, key):
        if self._envoy._tracer is None:
            return getattr(self._model, key)
        return super().__getattr__(key)

    def __setattr__(self, key, value):
        if (
            (key not in ("_model", "_model_key"))
            and (isinstance(value, torch.nn.Module))
            and (self._envoy._tracer is None)
        ):
            setattr(self._model, key, value)
        else:
            super().__setattr__(key, value)

    @property
    def device(self):
        """Returns the device."""
        return next(self.parameters()).device

    @device.setter
    def device(self, device: torch.device):
        """Sets the device."""
        self.to(device)

    @classmethod
    def from_path(cls, model_path: str) -> "LczeroModel":
        """Creates a wrapper from a model path.

        Parameters
        ----------
        model_path : str
            Path to the model file (.onnx or .pt)

        Returns
        -------
        LczeroModel
            The wrapped model instance

        Raises
        ------
        NotImplementedError
            If the model file extension is not supported
        """
        if model_path.endswith(".onnx"):
            return cls.from_onnx_path(model_path)
        elif model_path.endswith(".pt"):
            return cls.from_torch_path(model_path)
        else:
            raise NotImplementedError(f"Model path {model_path} is not supported.")

    @classmethod
    def from_onnx_path(cls, onnx_model_path: str, check: bool = True) -> "LczeroModel":
        """Builds a model from an ONNX file path.

        Parameters
        ----------
        onnx_model_path : str
            Path to the ONNX model file
        check : bool, optional
            Whether to perform shape inference check, by default True

        Returns
        -------
        LczeroModel
            The wrapped model instance

        Raises
        ------
        FileNotFoundError
            If the model file does not exist
        ValueError
            If the model could not be loaded
        """
        if not os.path.exists(onnx_model_path):
            raise FileNotFoundError(f"Model path {onnx_model_path} does not exist.")
        try:
            if check:
                onnx_model = safe_shape_inference(onnx_model_path)
            onnx_torch_model = convert(onnx_model)
            return cls(onnx_torch_model)
        except Exception as e:
            raise ValueError(f"Could not load model at {onnx_model_path}.") from e

    @classmethod
    def from_torch_path(cls, torch_model_path: str) -> "LczeroModel":
        """Builds a model from a PyTorch file path.

        Parameters
        ----------
        torch_model_path : str
            Path to the PyTorch model file

        Returns
        -------
        LczeroModel
            The wrapped model instance

        Raises
        ------
        FileNotFoundError
            If the model file does not exist
        ValueError
            If the model could not be loaded or is not a valid model type
        """
        if not os.path.exists(torch_model_path):
            raise FileNotFoundError(f"Model path {torch_model_path} does not exist.")
        try:
            torch_model = torch.load(torch_model_path)
        except Exception as e:
            raise ValueError(f"Could not load model at {torch_model_path}.") from e
        if isinstance(torch_model, LczeroModel):
            return torch_model
        elif isinstance(torch_model, nn.Module):
            return cls(torch_model)
        else:
            raise ValueError(f"Could not load model at {torch_model_path}.")

    @contextmanager
    def _ensure_proper_forward(self):
        old_forward = self._model.forward

        output_node = list(self._model.graph.nodes)[-1]
        output_names = [n.name.replace("output_", "") for n in output_node.all_input_nodes]

        def td_forward(x):
            old_out = old_forward(x)
            return TensorDict(
                {name: old_out[i] for i, name in enumerate(output_names)},
                batch_size=x.shape[0],
            )

        self._model.forward = td_forward
        yield
        self._model.forward = old_forward


class Flow(LczeroModel):
    """Base class for isolating a flow."""

    _flow_type: str
    _registry: Dict[str, Type["Flow"]] = {}

    def __init__(
        self,
        model_key,
        *args,
        **kwargs,
    ):
        if isinstance(model_key, LczeroModel):
            raise ValueError("Use the `from_model` classmethod to create a flow.")
        if not self.is_compatible(model_key):
            raise ValueError(f"The model does not have a {self._flow_type} head.")
        super().__init__(model_key, *args, **kwargs)

    @classmethod
    def register(cls, name: str):
        """Registers the flow.

        Parameters
        ----------
        name : str
            The name of the flow to register.

        Returns
        -------
        Callable
            Decorator function that registers the flow subclass.

        Raises
        ------
        ValueError
            If the flow name is already registered.
        """

        if name in cls._registry:
            raise ValueError(f"Flow {name} already registered.")

        def decorator(subclass):
            cls._registry[name] = subclass
            subclass._flow_type = name
            return subclass

        return decorator

    @classmethod
    def from_name(cls, name: str, *args, **kwargs) -> "Flow":
        """Returns the flow from its name.

        Parameters
        ----------
        name : str
            The name of the flow to instantiate.
        *args
            Positional arguments passed to flow constructor.
        **kwargs
            Keyword arguments passed to flow constructor.

        Returns
        -------
        Flow
            The instantiated flow.

        Raises
        ------
        KeyError
            If the flow name is not found.
        """
        if name not in cls._registry:
            raise KeyError(f"Flow {name} not found.")
        return cls._registry[name](*args, **kwargs)

    @classmethod
    def from_model(cls, name: str, model: LczeroModel, *args, **kwargs) -> "Flow":
        """Returns the flow from a model.

        Parameters
        ----------
        name : str
            The name of the flow to instantiate.
        model : LczeroModel
            The model to create the flow from.
        *args
            Positional arguments passed to flow constructor.
        **kwargs
            Keyword arguments passed to flow constructor.

        Returns
        -------
        Flow
            The instantiated flow.

        Raises
        ------
        KeyError
            If the flow name is not found.
        """
        if name not in cls._registry:
            raise KeyError(f"Flow {name} not found.")
        flow_class = cls._registry[name]
        return flow_class(model._model, *args, **kwargs)

    @classmethod
    def is_compatible(cls, model: nn.Module) -> bool:
        """Checks if the model is compatible with this flow.

        Parameters
        ----------
        model : nn.Module
            The model to check compatibility with.

        Returns
        -------
        bool
            Whether the model is compatible with this flow.
        """
        return hasattr(model, cls._flow_type) or hasattr(model, f"output/{cls._flow_type}")

    @contextmanager
    def _ensure_proper_forward(self):
        """Rewrites the forward function to return the flow output."""
        flow_type = getattr(self, "_flow_type", None)
        if flow_type is None:
            return

        with super()._ensure_proper_forward():
            old_forward = self._model.forward

            def flow_forward(*inputs, **kwargs):
                out = old_forward(*inputs, **kwargs)
                return out[flow_type]

            self._model.forward = flow_forward
            yield
            self._model.forward = old_forward


@Flow.register("policy")
class PolicyFlow(Flow):
    """Class for isolating the policy flow."""


@Flow.register("value")
class ValueFlow(Flow):
    """Class for isolating the value flow."""


@Flow.register("wdl")
class WdlFlow(Flow):
    """Class for isolating the WDL flow."""


@Flow.register("mlh")
class MlhFlow(Flow):
    """Class for isolating the MLH flow."""


@Flow.register("force_value")
class ForceValueFlow(Flow):
    """Class for forcing and isolating the value flow."""

    @classmethod
    def is_compatible(cls, model: nn.Module):
        return ValueFlow.is_compatible(model) or WdlFlow.is_compatible(model)

    @contextmanager
    def _ensure_proper_forward(self):
        flow_type = getattr(self, "_flow_type", None)
        if flow_type is None:
            return

        with LczeroModel._ensure_proper_forward(self):
            old_forward = self._model.forward

            def flow_forward(*inputs, **kwargs):
                out = old_forward(*inputs, **kwargs)
                if "value" in out.keys():
                    return out["value"]
                return out["wdl"] @ torch.tensor([1.0, 0.0, -1.0], device=out.device)

            self._model.forward = flow_forward
            yield
            self._model.forward = old_forward



"""Activation lens for XAI."""

from typing import Any, Optional, Callable
from dataclasses import dataclass

import torch
from datasets import Dataset
from torch.utils.data import DataLoader, TensorDataset

from lczerolens.model import LczeroModel


@dataclass
class ActivationBuffer:
    model: LczeroModel
    dataset: Dataset
    compute_fn: Callable[[Any, LczeroModel], torch.Tensor]
    n_batches_in_buffer: int = 10
    compute_batch_size: int = 64
    train_batch_size: int = 2048
    dataloader_kwargs: Optional[dict] = None
    logger: Optional[Callable] = None

    def __post_init__(self):
        if self.dataloader_kwargs is None:
            self.dataloader_kwargs = {}
        self._buffer = []
        self._remainder = None
        self._make_dataloader_it()

    def _make_dataloader_it(self):
        self._dataloader_it = iter(
            DataLoader(self.dataset, batch_size=self.compute_batch_size, **self.dataloader_kwargs)
        )

    @torch.no_grad
    def _fill_buffer(self):
        if self.logger is not None:
            self.logger.info("Computing activations...")
        self._buffer = []
        while len(self._buffer) < self.n_batches_in_buffer:
            try:
                next_batch = next(self._dataloader_it)
            except StopIteration:
                break
            activations = self.compute_fn(next_batch, self.model)
            self._buffer.append(activations.to("cpu"))
        if not self._buffer:
            raise StopIteration

    def _make_activations_it(self):
        if self._remainder is not None:
            self._buffer.append(self._remainder)
            self._remainder = None
        activations_ds = TensorDataset(torch.cat(self._buffer, dim=0))
        if self.logger is not None:
            self.logger.info(f"Activations dataset of size {len(activations_ds)}")

        self._activations_it = iter(
            DataLoader(
                activations_ds,
                batch_size=self.train_batch_size,
                shuffle=True,
            )
        )

    def __iter__(self):
        self._make_dataloader_it()
        self._fill_buffer()
        self._make_activations_it()
        self._remainder = None
        return self

    def __next__(self):
        try:
            activations = next(self._activations_it)[0]
            if activations.shape[0] < self.train_batch_size:
                self._remainder = activations
                self._fill_buffer()
                self._make_activations_it()
                activations = next(self._activations_it)[0]
            return activations
        except StopIteration:
            try:
                self._fill_buffer()
                self._make_activations_it()
                self.__next__()
            except StopIteration as e:
                if self._remainder is not None:
                    activations = self._remainder
                    self._remainder = None
                    return activations
                raise StopIteration from e
        raise StopIteration



"""Function classes to apply the LRP rules to the layers of the network.

Classes
-------
ElementwiseMultiplyUniform
    Distribute the relevance 100% to the input
SoftmaxEpsilon
    Softmax with epsilon.
MatrixMultiplicationEpsilon
    Matrix multiplication with epsilon.
"""

import torch
import torch.nn.functional as F
from torch.autograd import Function


def stabilize(tensor, epsilon=1e-6):
    return tensor + epsilon * ((-1) ** (tensor < 0))


class AddEpsilonFunction(Function):
    @staticmethod
    def forward(ctx, input_a, input_b, epsilon=1e-6):
        output = input_a + input_b
        ctx.save_for_backward(input_a, input_b, output, torch.tensor(epsilon))
        return output

    @staticmethod
    def backward(ctx, *grad_output):
        input_a, input_b, output, epsilon = ctx.saved_tensors
        out_relevance = grad_output[0] / stabilize(output, epsilon)
        return out_relevance * input_a, out_relevance * input_b, None


class AddEpsilon(torch.nn.Module):
    def __init__(self, epsilon=1e-6):
        super().__init__()
        self.epsilon = epsilon

    def forward(self, x, y):
        return AddEpsilonFunction.apply(x, y, self.epsilon)


class MatMulEpsilonFunction(Function):
    @staticmethod
    def forward(ctx, input, param, epsilon=1e-6):
        output = torch.matmul(input, param)
        ctx.save_for_backward(input, param, output, torch.tensor(epsilon))

        return output

    @staticmethod
    def backward(ctx, *grad_outputs):
        input, param, output, epsilon = ctx.saved_tensors
        out_relevance = grad_outputs[0]

        out_relevance = out_relevance / stabilize(output, epsilon)
        relevance = (out_relevance @ param.T) * input
        return relevance, None, None


class MatMulEpsilon(torch.nn.Module):
    def __init__(self, epsilon=1e-6):
        super().__init__()
        self.epsilon = epsilon

    def forward(self, x, y):
        return MatMulEpsilonFunction.apply(x, y, self.epsilon)


class BilinearMatMulEpsilonFunction(Function):
    @staticmethod
    def forward(ctx, input_a, input_b, epsilon=1e-6):
        outputs = torch.matmul(input_a, input_b)
        ctx.save_for_backward(input_a, input_b, outputs, torch.tensor(epsilon))

        return outputs

    @staticmethod
    def backward(ctx, *grad_outputs):
        input_a, input_b, outputs, epsilon = ctx.saved_tensors
        out_relevance = grad_outputs[0]

        out_relevance = out_relevance / stabilize(2 * outputs, epsilon)

        relevance_a = torch.matmul(out_relevance, input_b.permute(0, 1, -1, -2)) * input_a
        relevance_b = torch.matmul(input_a.permute(0, 1, -1, -2), out_relevance) * input_b

        return relevance_a, relevance_b, None


class BilinearMatMulEpsilon(torch.nn.Module):
    def __init__(self, epsilon=1e-6):
        super().__init__()
        self.epsilon = epsilon

    def forward(self, x, y):
        return BilinearMatMulEpsilonFunction.apply(x, y, self.epsilon)


class MulUniformFunction(Function):
    @staticmethod
    def forward(ctx, input_a, input_b):
        return input_a * input_b

    @staticmethod
    def backward(ctx, *grad_outputs):
        relevance = grad_outputs[0] * 0.5

        return relevance, relevance


class MulUniform(torch.nn.Module):
    def forward(self, x, y):
        return MulUniformFunction.apply(x, y)


class SoftmaxEpsilonFunction(Function):
    @staticmethod
    def forward(ctx, inputs, dim):
        outputs = F.softmax(inputs, dim=dim)
        ctx.save_for_backward(inputs, outputs)

        return outputs

    @staticmethod
    def backward(ctx, *grad_outputs):
        inputs, output = ctx.saved_tensors

        relevance = (grad_outputs[0] - (output * grad_outputs[0].sum(-1, keepdim=True))) * inputs

        return (relevance, None)



"""All concepts related to move."""

import chess
import torch

from lczerolens.board import LczeroBoard
from lczerolens.model import LczeroModel, PolicyFlow
from lczerolens.concept import BinaryConcept, MulticlassConcept


class BestLegalMove(MulticlassConcept):
    """Class for move concept-based XAI methods."""

    def __init__(
        self,
        model: LczeroModel,
    ):
        """Initialize the class."""
        self.policy_flow = PolicyFlow(model)

    def compute_label(
        self,
        board: LczeroBoard,
    ) -> int:
        """Compute the label for a given model and input."""
        (policy,) = self.policy_flow(board)
        policy = torch.softmax(policy.squeeze(0), dim=-1)

        legal_move_indices = [LczeroBoard.encode_move(move, board.turn) for move in board.legal_moves]
        sub_index = policy[legal_move_indices].argmax().item()
        return legal_move_indices[sub_index]


class PieceBestLegalMove(BinaryConcept):
    """Class for move concept-based XAI methods."""

    def __init__(
        self,
        model: LczeroModel,
        piece: str,
    ):
        """Initialize the class."""
        self.policy_flow = PolicyFlow(model)
        self.piece = chess.Piece.from_symbol(piece)

    def compute_label(
        self,
        board: LczeroBoard,
    ) -> int:
        """Compute the label for a given model and input."""
        (policy,) = self.policy_flow(board)
        policy = torch.softmax(policy.squeeze(0), dim=-1)

        legal_moves = list(board.legal_moves)
        legal_move_indices = [LczeroBoard.encode_move(move, board.turn) for move in legal_moves]
        sub_index = policy[legal_move_indices].argmax().item()
        best_legal_move = legal_moves[sub_index]
        if board.piece_at(best_legal_move.from_square) == self.piece:
            return 1
        return 0



"""Classes for playing."""

from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Optional, Callable, Tuple, Dict, Iterable

import chess
import torch
from torch.distributions import Categorical
from itertools import tee

from lczerolens.model import LczeroModel
from lczerolens.board import LczeroBoard


class Sampler(ABC):
    @abstractmethod
    def get_utilities(
        self, boards: Iterable[LczeroBoard], **kwargs
    ) -> Iterable[Tuple[torch.Tensor, torch.Tensor, Dict[str, float]]]:
        """Get the utility of the board.

        Parameters
        ----------
        boards : Iterable[LczeroBoard]
            The boards to evaluate.

        Returns
        -------
        Iterable[Tuple[torch.Tensor, torch.Tensor, Dict[str, float]]]
            The iterable over utilities, legal indices, and log dictionaries.
        """
        pass

    def choose_move(self, board: LczeroBoard, utility: torch.Tensor, legal_indices: torch.Tensor) -> chess.Move:
        """Choose the next moves.

        Parameters
        ----------
        board : LczeroBoard
            The board.
        utility : torch.Tensor
            The utility of the board.
        legal_indices : torch.Tensor
            The legal indices.

        Returns
        -------
        Iterable[chess.Move]
            The iterable over the moves.
        """
        m = Categorical(logits=utility)
        idx = m.sample()
        return board.decode_move(legal_indices[idx])

    def get_next_moves(self, boards: Iterable[LczeroBoard], **kwargs) -> Iterable[Tuple[chess.Move, Dict[str, float]]]:
        """Get the next move.

        Parameters
        ----------
        boards : Iterable[LczeroBoard]
            The boards to evaluate.

        Returns
        -------
        Iterable[Tuple[chess.Move, Dict[str, float]]]
            The iterable over the moves and log dictionaries.
        """
        util_boards, move_boards = tee(boards)
        for board, (utility, legal_indices, to_log) in zip(move_boards, self.get_utilities(util_boards, **kwargs)):
            predicted_move = self.choose_move(board, utility, legal_indices)
            yield predicted_move, to_log


class RandomSampler(Sampler):
    def get_utilities(
        self, boards: Iterable[LczeroBoard], **kwargs
    ) -> Iterable[Tuple[torch.Tensor, torch.Tensor, Dict[str, float]]]:
        for board in boards:
            legal_indices = board.get_legal_indices()
            utilities = torch.ones_like(legal_indices, dtype=torch.float32)
            yield utilities, legal_indices, {}


@dataclass
class ModelSampler(Sampler):
    model: LczeroModel
    use_argmax: bool = True
    alpha: float = 1.0
    beta: float = 1.0
    gamma: float = 1.0
    draw_score: float = 0.0
    m_max: float = 0.0345
    m_slope: float = 0.0027
    k_0: float = 0.0
    k_1: float = 1.6521
    k_2: float = -0.6521
    q_threshold: float = 0.8

    def choose_move(self, board: LczeroBoard, utility: torch.Tensor, legal_indices: torch.Tensor) -> chess.Move:
        if self.use_argmax:
            idx = utility.argmax()
            return board.decode_move(legal_indices[idx])
        return super().choose_move(board, utility, legal_indices)

    @torch.no_grad
    def get_utilities(
        self, boards: Iterable[LczeroBoard], **kwargs
    ) -> Iterable[Tuple[torch.Tensor, torch.Tensor, Dict[str, float]]]:
        batch_size = kwargs.pop("batch_size", -1)

        for legal_indices, batch_stats in self._get_batched_stats(boards, batch_size, **kwargs):
            to_log = {}
            utility = 0
            q_values = self._get_q_values(batch_stats, to_log)
            utility += self.alpha * q_values
            utility += self.beta * self._get_m_values(batch_stats, q_values, to_log)
            utility += self.gamma * self._get_p_values(batch_stats, legal_indices, to_log)
            to_log["max_utility"] = utility.max().item()
            yield utility, legal_indices, to_log

    def _get_batched_stats(self, boards, batch_size, use_next_boards=True, **kwargs):
        next_batch = []
        next_legal_indices = []

        def generator(next_batch, next_legal_indices):
            all_stats = self.model(*next_batch, **kwargs)
            offset = 0
            for legal_indices in next_legal_indices:
                n_boards = legal_indices.shape[0] + 1 if use_next_boards else 1
                batch_stats = all_stats[offset : offset + n_boards]
                offset += n_boards
                yield legal_indices, batch_stats

        for board in boards:
            legal_indices = board.get_legal_indices()
            next_boards = list(board.get_next_legal_boards()) if use_next_boards else []
            if len(next_batch) + len(next_boards) + 1 > batch_size and batch_size != -1:
                yield from generator(next_batch, next_legal_indices)
                next_batch = []
                next_legal_indices = []
            next_batch.extend([board] + next_boards)
            next_legal_indices.append(legal_indices)
        if next_batch:
            yield from generator(next_batch, next_legal_indices)

    def _get_q_values(self, batch_stats, to_log):
        if "value" in batch_stats.keys():
            to_log["value"] = batch_stats["value"][0].item()
            return batch_stats["value"][1:, 0]
        elif "wdl" in batch_stats.keys():
            to_log["wdl_w"] = batch_stats["wdl"][0][0].item()
            to_log["wdl_d"] = batch_stats["wdl"][0][1].item()
            to_log["wdl_l"] = batch_stats["wdl"][0][2].item()
            scores = torch.tensor([1, self.draw_score, -1])
            return batch_stats["wdl"][1:] @ scores
        return torch.zeros(batch_stats.batch_size[0] - 1)

    def _get_m_values(self, batch_stats, q_values, to_log):
        if "mlh" in batch_stats.keys():
            to_log["mlh"] = batch_stats["mlh"][0].item()
            delta_m_values = self.m_slope * (batch_stats["mlh"][1:, 0] - batch_stats["mlh"][0, 0])
            delta_m_values.clamp_(-self.m_max, self.m_max)
            scaled_q_values = torch.relu(q_values.abs() - self.q_threshold) / (1 - self.q_threshold)
            poly_q_values = self.k_0 + self.k_1 * scaled_q_values + self.k_2 * scaled_q_values**2
            return -q_values.sign() * delta_m_values * poly_q_values
        return torch.zeros(batch_stats.batch_size[0] - 1)

    def _get_p_values(
        self,
        batch_stats,
        legal_indices,
        to_log,
    ):
        if "policy" in batch_stats.keys():
            legal_policy = batch_stats["policy"][0].gather(0, legal_indices)
            to_log["max_legal_policy"] = legal_policy.max().item()
            return legal_policy
        return torch.zeros_like(legal_indices)


@dataclass
class PolicySampler(ModelSampler):
    use_suboptimal: bool = False

    @torch.no_grad
    def get_utilities(
        self, boards: Iterable[LczeroBoard], **kwargs
    ) -> Iterable[Tuple[torch.Tensor, torch.Tensor, Dict[str, float]]]:
        batch_size = kwargs.pop("batch_size", -1)

        to_log = {}
        for legal_indices, batch_stats in self._get_batched_stats(boards, batch_size, use_next_boards=False, **kwargs):
            legal_policy = batch_stats["policy"][0].gather(0, legal_indices.to(batch_stats["policy"].device))
            if self.use_suboptimal:
                idx = legal_policy.argmax()
                legal_policy[idx] = torch.tensor(-1e3)
            yield legal_policy, legal_indices, to_log


@dataclass
class SelfPlay:
    """A class for generating games."""

    white: Sampler
    black: Sampler

    def play(
        self,
        board: Optional[LczeroBoard] = None,
        max_moves: int = 100,
        to_play: chess.Color = chess.WHITE,
        report_fn: Optional[Callable[[dict, chess.Color], None]] = None,
        white_kwargs: Optional[Dict] = None,
        black_kwargs: Optional[Dict] = None,
    ):
        """
        Plays a game.
        """
        if board is None:
            board = LczeroBoard()
        if white_kwargs is None:
            white_kwargs = {}
        if black_kwargs is None:
            black_kwargs = {}
        game = []
        if to_play == chess.BLACK:
            move, _ = next(iter(self.black.get_next_moves([board], **black_kwargs)))
            board.push(move)
            game.append(move)
        for _ in range(max_moves):
            if board.is_game_over() or len(game) >= max_moves:
                break
            move, to_log = next(iter(self.white.get_next_moves([board], **white_kwargs)))
            if report_fn is not None:
                report_fn(to_log, board.turn)
            board.push(move)
            game.append(move)

            if board.is_game_over() or len(game) >= max_moves:
                break
            move, to_log = next(iter(self.black.get_next_moves([board], **black_kwargs)))
            if report_fn is not None:
                report_fn(to_log, board.turn)
            board.push(move)
            game.append(move)
            if board.is_game_over() or len(game) >= max_moves:
                break
        return game, board



"""Patching lens."""

from typing import Callable

from lczerolens.model import LczeroModel
from lczerolens.lens import Lens


@Lens.register("patching")
class PatchingLens(Lens):
    """
    Class for activation-based XAI methods.

    Examples
    --------

        .. code-block:: python

            model = LczeroModel.from_path(model_path)
            lens = PatchingLens()
            board = LczeroBoard()
            patch_fn = lambda n, m, *kwargs: pass
            results = lens.analyse(board, model=model)
    """

    def __init__(self, patch_fn: Callable, **kwargs):
        self._patch_fn = patch_fn
        super().__init__(**kwargs)

    def _intervene(
        self,
        model: LczeroModel,
        **kwargs,
    ) -> dict:
        for name, module in self._get_modules(model):
            self._patch_fn(name, module, **kwargs)
        return {}



"""Probing lens for XAI."""

from abc import ABC, abstractmethod
from typing import Any

import einops
import torch


EPS = 1e-6


class Probe(ABC):
    """Abstract class for probes."""

    def __init__(self):
        self._trained = False

    @abstractmethod
    def train(
        self,
        activations: torch.Tensor,
        labels: Any,
        **kwargs,
    ):
        """Train the probe."""
        pass

    @abstractmethod
    def predict(self, activations: torch.Tensor, **kwargs):
        """Predict with the probe."""
        pass


class SignalCav(Probe):
    """Signal CAV probe."""

    def train(
        self,
        activations: torch.Tensor,
        labels: torch.Tensor,
        **kwargs,
    ):
        if len(activations) != len(labels):
            raise ValueError("Number of activations and labels must match")
        if len(activations.shape) != 2:
            raise ValueError("Activations must a batch of tensors")
        if len(labels.shape) != 2:
            raise ValueError("Labels must a batch of tensors")

        mean_activation = activations.mean(dim=1, keepdim=True)
        mean_label = labels.mean(dim=1, keepdim=True)
        scaled_activations = activations - mean_activation
        scaled_labels = labels - mean_label
        cav = einops.einsum(scaled_activations, scaled_labels, "b a, b d -> a d")
        self._h = cav / (cav.norm(dim=0, keepdim=True) + EPS)

    def predict(self, activations: torch.Tensor, **kwargs):
        if not self._trained:
            raise ValueError("Probe not trained")

        if len(activations.shape) != 2:
            raise ValueError("Activations must a batch of tensors")

        dot_prod = einops.einsum(activations, self._h, "b a, a d -> b d")
        return dot_prod / (activations.norm(dim=1, keepdim=True) + EPS)



"""
Play module for the lczerolens package.
"""

from .sampling import ModelSampler, Sampler, RandomSampler, PolicySampler
from .game import Game
from .puzzle import Puzzle
from . import sampling, game, puzzle

__all__ = ["ModelSampler", "Sampler", "RandomSampler", "PolicySampler", "Game", "Puzzle", "sampling", "game", "puzzle"]



"""Probing lens."""

from typing import Callable

from lczerolens.model import LczeroModel
from lczerolens.lens import Lens


@Lens.register("probing")
class ProbingLens(Lens):
    """
    Class for probing-based XAI methods.

    Examples
    --------

        .. code-block:: python

            model = LczeroModel.from_path(model_path)
            lens = ProbingLens(probe)
            board = LczeroBoard()
            results = lens.analyse(board, model=model)
    """

    def __init__(self, probe_fn: Callable, **kwargs):
        self._probe_fn = probe_fn
        super().__init__(**kwargs)

    def _intervene(
        self,
        model: LczeroModel,
        **kwargs,
    ) -> dict:
        return {name: self._probe_fn(module.output.save()) for name, module in self._get_modules(model)}



"""Compute LRP heatmap for a given model and input."""

from typing import Any


from lczerolens.model import LczeroModel
from lczerolens.lens import Lens


@Lens.register("lrp")
class LrpLens(Lens):
    """Class for wrapping the LCZero models."""

    def _intervene(
        self,
        model: LczeroModel,
        **kwargs,
    ) -> Any:
        # TODO: Refactor this logic
        pass



"""Class for concept-based XAI methods."""

from abc import ABC, abstractmethod
from typing import Any

import torch
from sklearn import metrics
from datasets import Features, Value, Sequence, ClassLabel

from lczerolens.board import LczeroBoard


class Concept(ABC):
    """
    Class for concept-based XAI methods.
    """

    @abstractmethod
    def compute_label(
        self,
        board: LczeroBoard,
    ) -> Any:
        """
        Compute the label for a given model and input.
        """
        pass

    @staticmethod
    @abstractmethod
    def compute_metrics(
        predictions,
        labels,
    ):
        """
        Compute the metrics for a given model and input.
        """
        pass

    @property
    @abstractmethod
    def features(self) -> Features:
        """
        Return the features for the concept.
        """
        pass


class BinaryConcept(Concept):
    """
    Class for binary concept-based XAI methods.
    """

    features = Features(
        {
            "gameid": Value("string"),
            "moves": Sequence(Value("string")),
            "fen": Value("string"),
            "label": ClassLabel(num_classes=2),
        }
    )

    @staticmethod
    def compute_metrics(
        predictions,
        labels,
    ):
        """
        Compute the metrics for a given model and input.
        """
        return {
            "accuracy": metrics.accuracy_score(labels, predictions),
            "precision": metrics.precision_score(labels, predictions),
            "recall": metrics.recall_score(labels, predictions),
            "f1": metrics.f1_score(labels, predictions),
        }


class NullConcept(BinaryConcept):
    """
    Class for binary concept-based XAI methods.
    """

    def compute_label(
        self,
        board: LczeroBoard,
    ) -> Any:
        """
        Compute the label for a given model and input.
        """
        return 0


class OrBinaryConcept(BinaryConcept):
    """
    Class for binary concept-based XAI methods.
    """

    def __init__(self, *concepts: BinaryConcept):
        for concept in concepts:
            if not isinstance(concept, BinaryConcept):
                raise ValueError(f"{concept} is not a BinaryConcept")
        self.concepts = concepts

    def compute_label(
        self,
        board: LczeroBoard,
    ) -> Any:
        """
        Compute the label for a given model and input.
        """
        return any(concept.compute_label(board) for concept in self.concepts)


class AndBinaryConcept(BinaryConcept):
    """
    Class for binary concept-based XAI methods.
    """

    def __init__(self, *concepts: BinaryConcept):
        for concept in concepts:
            if not isinstance(concept, BinaryConcept):
                raise ValueError(f"{concept} is not a BinaryConcept")
        self.concepts = concepts

    def compute_label(
        self,
        board: LczeroBoard,
    ) -> Any:
        """
        Compute the label for a given model and input.
        """
        return all(concept.compute_label(board) for concept in self.concepts)


class MulticlassConcept(Concept):
    """
    Class for multiclass concept-based XAI methods.
    """

    features = Features(
        {
            "gameid": Value("string"),
            "moves": Sequence(Value("string")),
            "fen": Value("string"),
            "label": Value("int32"),
        }
    )

    @staticmethod
    def compute_metrics(
        predictions,
        labels,
    ):
        """
        Compute the metrics for a given model and input.
        """
        return {
            "accuracy": metrics.accuracy_score(labels, predictions),
            "precision": metrics.precision_score(labels, predictions, average="weighted"),
            "recall": metrics.recall_score(labels, predictions, average="weighted"),
            "f1": metrics.f1_score(labels, predictions, average="weighted"),
        }


class ContinuousConcept(Concept):
    """
    Class for continuous concept-based XAI methods.
    """

    features = Features(
        {
            "gameid": Value("string"),
            "moves": Sequence(Value("string")),
            "fen": Value("string"),
            "label": Value("float32"),
        }
    )

    @staticmethod
    def compute_metrics(
        predictions,
        labels,
    ):
        """
        Compute the metrics for a given model and input.
        """
        return {
            "rmse": metrics.root_mean_squared_error(labels, predictions),
            "mae": metrics.mean_absolute_error(labels, predictions),
            "r2": metrics.r2_score(labels, predictions),
        }


def concept_collate_fn(batch):
    boards = []
    labels = []
    for element in batch:
        board = LczeroBoard(element["fen"])
        for move in element["moves"]:
            board.push_san(move)
        boards.append(board)
        labels.append(element["label"])
    return boards, labels, batch


def concept_init_rel(output, infos):
    labels = infos[0]
    rel = torch.zeros_like(output)
    for i in range(rel.shape[0]):
        rel[i, labels[i]] = output[i, labels[i]]
    return rel



"""Board class."""

import re
from enum import Enum
from typing import Optional, Generator, Tuple, List, Union, Any

import chess
import chess.svg
import torch
import io
import numpy as np

from .constants import INVERTED_POLICY_INDEX, POLICY_INDEX


class InputEncoding(int, Enum):
    """Input encoding for the board tensor."""

    INPUT_CLASSICAL_112_PLANE = 0
    INPUT_CLASSICAL_112_PLANE_REPEATED = 1
    INPUT_CLASSICAL_112_PLANE_NO_HISTORY = 2


class LczeroBoard(chess.Board):
    """A class for wrapping the LczeroBoard class."""

    @staticmethod
    def get_plane_order(us: bool):
        """Get the plane order for the given us view.

        Parameters
        ----------
        us : bool
            The us_them tuple.

        Returns
        -------
        str
            The plane order.
        """
        plane_orders = {chess.WHITE: "PNBRQK", chess.BLACK: "pnbrqk"}
        return plane_orders[us] + plane_orders[not us]

    @staticmethod
    def get_piece_index(piece: str, us: bool, plane_order: Optional[str] = None):
        """Converts a piece to its index in the plane order.

        Parameters
        ----------
        piece : str
            The piece to convert.
        us : bool
            The us_them tuple.
        plane_order : Optional[str]
            The plane order.

        Returns
        -------
        int
            The index of the piece in the plane order.
        """
        if plane_order is None:
            plane_order = LczeroBoard.get_plane_order(us)
        return f"{plane_order}0".index(piece)

    def to_config_tensor(
        self,
        us: Optional[bool] = None,
        input_encoding: InputEncoding = InputEncoding.INPUT_CLASSICAL_112_PLANE,
    ):
        """Converts a LczeroBoard to a tensor based on the pieces configuration.

        Parameters
        ----------
        us : Optional[bool]
            The us_them tuple.
        input_encoding : InputEncoding
            The input encoding method.

        Returns
        -------
        torch.Tensor
            The 13x8x8 tensor.
        """
        if not isinstance(input_encoding, InputEncoding):
            raise NotImplementedError(f"Input encoding {input_encoding} not implemented.")
        if us is None:
            us = self.turn
        plane_order = LczeroBoard.get_plane_order(us)

        def piece_to_index(piece: str):
            return f"{plane_order}0".index(piece)

        fen_board = self.fen().split(" ")[0]
        fen_rep = re.sub(r"(\d)", lambda m: "0" * int(m.group(1)), fen_board)
        rows = fen_rep.split("/")
        rev_rows = rows[::-1]
        ordered_fen = "".join(rev_rows)

        config_tensor = torch.zeros((13, 8, 8), dtype=torch.float)
        ordinal_board = torch.tensor(tuple(map(piece_to_index, ordered_fen)), dtype=torch.float)
        ordinal_board = ordinal_board.reshape((8, 8)).unsqueeze(0)
        piece_tensor = torch.tensor(tuple(map(piece_to_index, plane_order)), dtype=torch.float)
        piece_tensor = piece_tensor.reshape((12, 1, 1))
        config_tensor[:12] = (ordinal_board == piece_tensor).float()
        if self.is_repetition(2):  # Might be wrong if the full history is not available
            config_tensor[12] = torch.ones((8, 8), dtype=torch.float)
        return config_tensor if us == chess.WHITE else config_tensor.flip(1)

    def to_input_tensor(
        self,
        with_history: bool = True,
        input_encoding: InputEncoding = InputEncoding.INPUT_CLASSICAL_112_PLANE,
    ):
        """Create the lc0 input tensor from the history of a game.

        Parameters
        ----------
        with_history : bool
            Whether to include the history of the game.
        input_encoding : InputEncoding
            The input encoding method.

        Returns
        -------
        torch.Tensor
            The 112x8x8 tensor.
        """
        if not isinstance(input_encoding, InputEncoding):
            raise NotImplementedError(f"Input encoding {input_encoding} not implemented.")

        input_tensor = torch.zeros((112, 8, 8), dtype=torch.float)
        us = self.turn
        them = not us
        moves = []

        if with_history:
            if (
                input_encoding == InputEncoding.INPUT_CLASSICAL_112_PLANE
                or input_encoding == InputEncoding.INPUT_CLASSICAL_112_PLANE_REPEATED
            ):
                for i in range(8):
                    config_tensor = self.to_config_tensor(us)
                    input_tensor[i * 13 : (i + 1) * 13] = config_tensor
                    try:
                        moves.append(self.pop())
                    except IndexError:
                        if input_encoding == InputEncoding.INPUT_CLASSICAL_112_PLANE_REPEATED:
                            input_tensor[(i + 1) * 13 : 104] = config_tensor.repeat(7 - i, 1, 1)
                        break

            elif input_encoding == InputEncoding.INPUT_CLASSICAL_112_PLANE_NO_HISTORY:
                config_tensor = self.to_config_tensor(us)
                input_tensor[:104] = config_tensor.repeat(8, 1, 1)
            else:
                raise ValueError(f"Got unexpected input encoding {input_encoding}")

        # Restore the moves
        for move in reversed(moves):
            self.push(move)

        if self.has_queenside_castling_rights(us):
            input_tensor[104] = torch.ones((8, 8), dtype=torch.float)
        if self.has_kingside_castling_rights(us):
            input_tensor[105] = torch.ones((8, 8), dtype=torch.float)
        if self.has_queenside_castling_rights(them):
            input_tensor[106] = torch.ones((8, 8), dtype=torch.float)
        if self.has_kingside_castling_rights(them):
            input_tensor[107] = torch.ones((8, 8), dtype=torch.float)
        if us == chess.BLACK:
            input_tensor[108] = torch.ones((8, 8), dtype=torch.float)
        input_tensor[109] = torch.ones((8, 8), dtype=torch.float) * self.halfmove_clock
        input_tensor[111] = torch.ones((8, 8), dtype=torch.float)

        return input_tensor

    @staticmethod
    def encode_move(
        move: chess.Move,
        us: bool,
    ) -> int:
        """
        Converts a chess.Move object to an index.

        Parameters
        ----------
        move : chess.Move
            The chess move to encode.
        us : bool
            The side to move (True for white, False for black).

        Returns
        -------
        int
            The encoded move index.
        """
        from_square = move.from_square
        to_square = move.to_square

        if us == chess.BLACK:
            from_square_row = from_square // 8
            from_square_col = from_square % 8
            from_square = 8 * (7 - from_square_row) + from_square_col
            to_square_row = to_square // 8
            to_square_col = to_square % 8
            to_square = 8 * (7 - to_square_row) + to_square_col
        us_uci_move = chess.SQUARE_NAMES[from_square] + chess.SQUARE_NAMES[to_square]
        if move.promotion is not None:
            if move.promotion == chess.BISHOP:
                us_uci_move += "b"
            elif move.promotion == chess.ROOK:
                us_uci_move += "r"
            elif move.promotion == chess.QUEEN:
                us_uci_move += "q"
            # Knight promotion is the default
        return INVERTED_POLICY_INDEX[us_uci_move]

    def decode_move(
        self,
        index: int,
    ) -> chess.Move:
        """
        Converts an index to a chess.Move object.

        Parameters
        ----------
        index : int
            The index to convert.

        Returns
        -------
        chess.Move
            The chess move.
        """
        us = self.turn
        us_uci_move = POLICY_INDEX[index]
        from_square = chess.SQUARE_NAMES.index(us_uci_move[:2])
        to_square = chess.SQUARE_NAMES.index(us_uci_move[2:4])
        if us == chess.BLACK:
            from_square_row = from_square // 8
            from_square_col = from_square % 8
            from_square = 8 * (7 - from_square_row) + from_square_col
            to_square_row = to_square // 8
            to_square_col = to_square % 8
            to_square = 8 * (7 - to_square_row) + to_square_col

        uci_move = chess.SQUARE_NAMES[from_square] + chess.SQUARE_NAMES[to_square]
        from_piece = self.piece_at(from_square)
        if from_piece == chess.PAWN and to_square >= 56:  # Knight promotion is the default
            uci_move += "n"
        return chess.Move.from_uci(uci_move)

    def get_legal_indices(
        self,
    ) -> torch.Tensor:
        """
        Gets the legal indices.

        Returns
        -------
        torch.Tensor
            Tensor containing indices of legal moves.
        """
        us = self.turn
        return torch.tensor([self.encode_move(move, us) for move in self.legal_moves])

    def get_next_legal_boards(
        self,
        n_history: int = 7,
    ) -> Generator["LczeroBoard", None, None]:
        """
        Gets the next legal boards.

        Parameters
        ----------
        n_history : int, optional
            Number of previous positions to keep in the move stack, by default 7.

        Returns
        -------
        Generator[LczeroBoard, None, None]
            Generator yielding board positions after each legal move.
        """
        working_board = self.copy(stack=n_history)
        for move in working_board.legal_moves:
            working_board.push(move)
            yield working_board.copy(stack=n_history)
            working_board.pop()

    def render_heatmap(
        self,
        heatmap: Union[torch.Tensor, np.ndarray],
        square: Optional[str] = None,
        vmin: Optional[float] = None,
        vmax: Optional[float] = None,
        arrows: Optional[List[Tuple[str, str]]] = None,
        normalise: str = "none",
        save_to: Optional[str] = None,
        cmap_name: str = "RdYlBu_r",
        alpha: float = 1.0,
        flip_mode: str = "board",
    ) -> Tuple[Optional[str], Any]:
        """Render a heatmap on the board.

        Parameters
        ----------
        heatmap : torch.Tensor or numpy.ndarray
            The heatmap values to visualize on the board (64,) or (8, 8).
        square : Optional[str], default=None
            Chess square to highlight (e.g. 'e4').
        vmin : Optional[float], default=None
            Minimum value for the colormap normalization.
        vmax : Optional[float], default=None
            Maximum value for the colormap normalization.
        arrows : Optional[List[Tuple[str, str]]], default=None
            List of arrow tuples (from_square, to_square) to draw on board.
        normalise : str, default="none"
            Normalization method. Use "abs" for absolute value normalization.
        save_to : Optional[str], default=None
            Path to save the visualization. If None, returns the figure.
        cmap_name : str, default="RdYlBu_r"
            Name of matplotlib colormap to use.
        alpha : float, default=1.0
            Opacity of the heatmap overlay.
        flip_mode : str, default="board"
            Flip mode for black's perspective. Use "board" to flip the board, "heatmap" to flip the heatmap.

        Returns
        -------
        Union[Tuple[str, matplotlib.figure.Figure], None]
            If save_to is None, returns (SVG string, matplotlib figure).
            If save_to is provided, saves files and returns None.

        Raises
        ------
        ValueError
            If save_to is provided and does not end with `.svg`.
        """
        try:
            import matplotlib
            import matplotlib.pyplot as plt
        except ImportError as e:
            raise ImportError(
                "matplotlib is required to render heatmaps, install it with `pip install lczerolens[viz]`."
            ) from e

        if flip_mode not in ["board", "heatmap"]:
            raise ValueError(f"Got unexpected flip_mode {flip_mode}")

        if heatmap.ndim == 2:
            heatmap = heatmap.view(64)
        if flip_mode == "heatmap":
            heatmap = heatmap.view(8, 8).flip(0).view(64)

        cmap = matplotlib.colormaps[cmap_name].resampled(1000)

        if normalise == "abs":
            a_max = heatmap.abs().max()
            if a_max != 0:
                heatmap = heatmap / a_max
            vmin = -1
            vmax = 1
        if vmin is None:
            vmin = heatmap.min()
        if vmax is None:
            vmax = heatmap.max()
        norm = matplotlib.colors.Normalize(vmin=vmin, vmax=vmax, clip=False)

        color_dict = {}
        for square_index in range(64):
            color = cmap(norm(heatmap[square_index]))
            color = (*color[:3], alpha)
            color_dict[square_index] = matplotlib.colors.to_hex(color, keep_alpha=True)
        fig = plt.figure(figsize=(1, 4.1))
        ax = plt.gca()
        ax.axis("off")
        fig.colorbar(
            matplotlib.cm.ScalarMappable(norm=norm, cmap=cmap),
            ax=ax,
            orientation="vertical",
            fraction=1.0,
        )
        if square is not None:
            try:
                check = chess.parse_square(square)
            except ValueError:
                check = None
        else:
            check = None
        if arrows is None:
            arrows = []

        svg_board = chess.svg.board(
            self,
            orientation=self.turn if flip_mode == "board" else chess.WHITE,
            check=check,
            fill=color_dict,
            size=400,
            arrows=arrows,
        )
        buffer = io.BytesIO()
        fig.savefig(buffer, format="svg")
        svg_colorbar = buffer.getvalue().decode("utf-8")
        plt.close()

        if save_to is None:
            return svg_board, svg_colorbar
        elif not save_to.endswith(".svg"):
            raise ValueError("only saving to `svg` is supported")

        with open(save_to.replace(".svg", "_board.svg"), "w") as f:
            f.write(svg_board)
        with open(save_to.replace(".svg", "_colorbar.svg"), "w") as f:
            f.write(svg_colorbar)



"""Preproces functions for chess puzzles."""

from dataclasses import dataclass
from typing import Dict, List, Union, Tuple, Optional, Iterable

import chess
import torch
from datasets import Features, Value
from itertools import tee, chain

from lczerolens.board import LczeroBoard
from .sampling import Sampler


PUZZLE_DATASET_FEATURES = Features(
    {
        "PuzzleId": Value("string"),
        "FEN": Value("string"),
        "Moves": Value("string"),
        "Rating": Value("int64"),
        "RatingDeviation": Value("int64"),
        "Popularity": Value("int64"),
        "NbPlays": Value("int64"),
        "Themes": Value("string"),
        "GameUrl": Value("string"),
        "OpeningTags": Value("string"),
    }
)


@dataclass
class Puzzle:
    puzzle_id: str
    fen: str
    initial_move: chess.Move
    moves: List[chess.Move]
    rating: int
    rating_deviation: int
    popularity: int
    nb_plays: int
    themes: List[str]
    game_url: str
    opening_tags: List[str]

    @classmethod
    def from_dict(cls, obj: Dict[str, Union[str, int, None]]) -> "Puzzle":
        uci_moves = obj["Moves"].split()
        moves = [chess.Move.from_uci(uci_move) for uci_move in uci_moves]
        return cls(
            puzzle_id=obj["PuzzleId"],
            fen=obj["FEN"],
            initial_move=moves[0],
            moves=moves[1:],
            rating=obj["Rating"],
            rating_deviation=obj["RatingDeviation"],
            popularity=obj["Popularity"],
            nb_plays=obj["NbPlays"],
            themes=obj["Themes"].split() if obj["Themes"] is not None else [],
            game_url=obj["GameUrl"],
            opening_tags=obj["OpeningTags"].split() if obj["OpeningTags"] is not None else [],
        )

    def __len__(self) -> int:
        return len(self.moves)

    @property
    def initial_board(self) -> LczeroBoard:
        board = LczeroBoard(self.fen)
        board.push(self.initial_move)
        return board

    def board_move_generator(self, all_moves: bool = False) -> Iterable[Tuple[LczeroBoard, chess.Move]]:
        board = self.initial_board
        initial_turn = board.turn
        for move in self.moves:
            if not all_moves and board.turn != initial_turn:
                board.push(move)
                continue
            yield board.copy(), move
            board.push(move)

    @classmethod
    def evaluate_multiple(
        cls,
        puzzles: Iterable["Puzzle"],
        sampler: Sampler,
        all_moves: bool = False,
        compute_metrics: bool = True,
        **kwargs,
    ) -> Union[Iterable[Dict[str, float]], Iterable[Tuple[torch.Tensor, torch.Tensor, chess.Move]]]:
        metric_puzzles, board_move_puzzles = tee(puzzles)
        board_move_generator = chain.from_iterable(
            puzzle.board_move_generator(all_moves) for puzzle in board_move_puzzles
        )

        def board_generator():
            for board, _ in board_move_generator:
                yield board

        util_boards, move_boards = tee(board_generator())

        def metric_inputs_generator():
            util_gen = sampler.get_utilities(util_boards, **kwargs)
            for board, (utility, legal_indices, _) in zip(move_boards, util_gen):
                predicted_move = sampler.choose_move(board, utility, legal_indices)
                yield utility, legal_indices, predicted_move

        if compute_metrics:
            return cls.compute_metrics(metric_puzzles, metric_inputs_generator(), all_moves=all_moves)
        else:
            return metric_inputs_generator()

    def evaluate(self, sampler: Sampler, all_moves: bool = False, **kwargs) -> Tuple[float, Optional[float]]:
        return next(iter(self.evaluate_multiple([self], sampler, all_moves, **kwargs)))

    @staticmethod
    def compute_metrics(
        puzzles: Iterable["Puzzle"],
        inputs: Iterable[Tuple[torch.Tensor, torch.Tensor, chess.Move]],
        all_moves: bool = False,
    ) -> Iterable[Dict[str, float]]:
        iter_inputs = iter(inputs)
        for puzzle in puzzles:
            total = len(puzzle) if all_moves else (len(puzzle) + 1) // 2
            metrics = {"score": 0.0, "perplexity": 1.0, "normalized_perplexity": 1.0}
            for board, move in puzzle.board_move_generator(all_moves=all_moves):
                utility, legal_indices, predicted_move = next(iter_inputs)
                index = LczeroBoard.encode_move(move, board.turn)
                probs = torch.softmax(utility, dim=0)
                move_prob = probs[legal_indices == index].item()
                metrics["perplexity"] *= move_prob ** (-1 / total)
                metrics["normalized_perplexity"] *= (len(legal_indices) * move_prob) ** (-1 / total)
                if predicted_move == move:
                    metrics["score"] += 1
            metrics["score"] /= total
            yield metrics

    def _repr_svg_(self) -> str:
        return self.initial_board._repr_svg_()



"""Preproces functions for chess games."""

from dataclasses import dataclass
from typing import Any, Dict, List, Optional, Union

from datasets import Features, Value, Sequence

from lczerolens.board import LczeroBoard

GAME_DATASET_FEATURES = Features(
    {
        "gameid": Value("string"),
        "moves": Value("string"),
    }
)

BOARD_DATASET_FEATURES = Features(
    {
        "gameid": Value("string"),
        "moves": Sequence(Value("string")),
        "fen": Value("string"),
    }
)


@dataclass
class Game:
    gameid: str
    moves: List[str]
    book_exit: Optional[int] = None

    @classmethod
    def from_dict(cls, obj: Dict[str, str]) -> "Game":
        if "moves" not in obj:
            ValueError("The dict should contain `moves`.")
        if "gameid" not in obj:
            ValueError("The dict should contain `gameid`.")
        *pre, post = obj["moves"].split("{ Book exit }")
        if pre:
            if len(pre) > 1:
                raise ValueError("More than one book exit")
            (pre,) = pre
            parsed_pre_moves = [m for m in pre.split() if not m.endswith(".")]
            book_exit = len(parsed_pre_moves)
        else:
            parsed_pre_moves = []
            book_exit = None
        parsed_moves = parsed_pre_moves + [m for m in post.split() if not m.endswith(".")]
        return cls(
            gameid=obj["gameid"],
            moves=parsed_moves,
            book_exit=book_exit,
        )

    def to_boards(
        self,
        n_history: int = 0,
        skip_book_exit: bool = False,
        skip_first_n: int = 0,
        output_dict=True,
    ) -> List[Union[Dict[str, Any], LczeroBoard]]:
        working_board = LczeroBoard()
        if skip_first_n > 0 or (skip_book_exit and (self.book_exit is not None)):
            boards = []
        else:
            if output_dict:
                boards = [
                    {
                        "fen": working_board.fen(),
                        "moves": [],
                        "gameid": self.gameid,
                    }
                ]
            else:
                boards = [working_board.copy(stack=n_history)]

        for i, move in enumerate(self.moves[:-1]):  # skip the last move as it can be over
            working_board.push_san(move)
            if (i < skip_first_n) or (skip_book_exit and (self.book_exit is not None) and (i < self.book_exit)):
                continue
            if output_dict:
                save_board = working_board.copy(stack=n_history)
                boards.append(
                    {
                        "fen": save_board.root().fen(),
                        "moves": [move.uci() for move in save_board.move_stack],
                        "gameid": self.gameid,
                    }
                )
            else:
                boards.append(working_board.copy(stack=n_history))
        return boards

    @staticmethod
    def board_collate_fn(batch):
        boards = []
        for element in batch:
            board = LczeroBoard(element["fen"])
            for move in element["moves"]:
                board.push_san(move)
            boards.append(board)
        return boards, {}



"""
Lenses module.
"""

from .activation import ActivationLens
from .composite import CompositeLens
from .gradient import GradientLens
from .lrp import LrpLens
from .patching import PatchingLens
from .probing import ProbingLens

__all__ = [
    "ActivationLens",
    "CompositeLens",
    "GradientLens",
    "LrpLens",
    "PatchingLens",
    "ProbingLens",
]



"""Compute Gradient heatmap for a given model and input."""

from lczerolens.model import LczeroModel
from lczerolens.lens import Lens


@Lens.register("gradient")
class GradientLens(Lens):
    """Class for gradient-based XAI methods."""

    def __init__(self, *, input_requires_grad: bool = True, **kwargs):
        self.input_requires_grad = input_requires_grad
        super().__init__(**kwargs)

    def _intervene(self, model: LczeroModel, **kwargs) -> dict:
        init_target = kwargs.get("init_target", lambda model: model.output["value"])
        init_gradient = kwargs.get("init_gradient", lambda model: None)

        results = {}
        if self.input_requires_grad:
            model.input.requires_grad_(self.input_requires_grad)
            results["input_grad"] = model.input.grad.save()
        for name, module in self._get_modules(model):
            results[f"{name}_output_grad"] = module.output.grad.save()
        target = init_target(model)
        target.backward(gradient=init_gradient(model))
        return results



"""Composite lens for XAI."""

from typing import List, Dict, Union, Any

from lczerolens.lens import Lens
from lczerolens.model import LczeroModel


class CompositeLens(Lens):
    """Composite lens for XAI.

    Examples
    --------

        .. code-block:: python

            model = LczeroModel.from_path(model_path)
            lens = CompositeLens([ActivationLens(), GradientLens()])
            board = LczeroBoard()
            results = lens.analyse(board, model=model)
    """

    def __init__(self, lenses: Union[List[Lens], Dict[str, Lens]], merge_results: bool = True):
        self._lens_map = lenses if isinstance(lenses, dict) else {f"lens_{i}": lens for i, lens in enumerate(lenses)}
        self.merge_results = merge_results

    def is_compatible(self, model: LczeroModel) -> bool:
        return all(lens.is_compatible(model) for lens in self._lens_map.values())

    def prepare(self, model: LczeroModel, **kwargs) -> LczeroModel:
        for lens in self._lens_map.values():
            model = lens.prepare(model, **kwargs)
        return model

    def _intervene(self, model: LczeroModel, **kwargs) -> Dict[str, Any]:
        results = {name: lens._intervene(model, **kwargs) for name, lens in self._lens_map.items()}
        if self.merge_results:
            return {k: v for d in results.values() for k, v in d.items()}
        return results



"""Main module for the lczerolens package."""

from importlib.metadata import PackageNotFoundError, version

from .board import LczeroBoard, InputEncoding
from .model import LczeroModel, Flow
from .lens import Lens
from . import lenses, concepts, play

try:
    __version__ = version("lczerolens")
except PackageNotFoundError:
    __version__ = "unknown version"

__all__ = [
    "LczeroBoard",
    "LczeroModel",
    "Flow",
    "InputEncoding",
    "Lens",
    "lenses",
    "concepts",
    "play",
]



"""Concepts module."""

from .material import HasMaterialAdvantage, HasPiece
from .move import BestLegalMove, PieceBestLegalMove
from .threat import HasMateThreat, HasThreat

__all__ = [
    "HasPiece",
    "HasThreat",
    "HasMateThreat",
    "HasMaterialAdvantage",
    "BestLegalMove",
    "PieceBestLegalMove",
]
